#include <stdio.h> // printf
#include <stdlib.h> // calloc
#include <assert.h> // assert
#include <cuda.h> // CUDA library
#include "helper_cuda.h" // CUDA error checking

// integer ceiling division
#define DIV_CEIL(a,b) (((a)+(b)-1)/(b))
// maximum
#define MIN(a,b) (((a)<(b))?(a):(b))


#define BLOCK_SIZE 256
#define NUM_WARPS (BLOCK_SIZE/CUDA_WARP_SIZE)


__global__ void set(const uint N, int *values) {
	int gtid = blockIdx.x*blockDim.x + threadIdx.x;
	if (gtid<N)
		values[gtid] = (int) CUDA_WARP_SIZE*cosf((float)gtid);
}

// To avoid bank conflicts, or not!
#define WARP_STRIDE (CUDA_WARP_SIZE)

__global__ void multi_scan(const int *values, int *inclusive, int *exclusive) {
	__shared__ volatile int scan[NUM_WARPS*WARP_STRIDE];
	int tid = threadIdx.x;
	int lane = tid & CUDA_WARP_MASK;
	int warp = tid / CUDA_WARP_SIZE;
	int gtid = blockIdx.x * blockDim.x + tid;

	// Read from global memory.
	int x = values[gtid];
	int s = warp*WARP_STRIDE+lane;
	scan[s] = x;

	// Run inclusive scan on each warp's data.
	int sum = x;
	#pragma unroll
	for (int offset=1; offset<CUDA_WARP_SIZE; offset*=2) {
		if (lane>=offset) sum += scan[s-offset];
		scan[s] = sum;
	}

	// Synchronize to make all the totals available to the reduction code.
	__syncthreads();

	// Reducition of the totals
	__shared__ volatile int totals[NUM_WARPS];
	if (tid < NUM_WARPS) {
		// Grab the block total for the tid'th block. This is the last element
		// in the block's scanned sequence.
		int total = scan[tid*WARP_STRIDE + CUDA_WARP_SIZE-1];
		totals[tid] = total;

		int totalsSum = total;
		#pragma unroll
		for(int offset=1; offset<NUM_WARPS; offset*=2) {
			if (tid>=offset) totalsSum += totals[tid-offset];
			totals[tid] = totalsSum;
		}

		// Subtract total from totalsSum for an exclusive scan.
		totals[tid] = totalsSum - total;
	}

	// Synchronize to make the block scan available to all warps.
	__syncthreads();

	// Add the block scan to the inclusive sum for the block.
	sum += totals[warp];

	// Write the inclusive and exclusive scans to global memory.
	inclusive[gtid] = sum;
	exclusive[gtid] = sum - x;
}


int main(int argc, char **argv)
{	
	int *d_values = NULL, *h_values = NULL; // pointers in Host and Device
	int *d_inclusive = NULL, *h_inclusive = NULL;
	int *d_exclusive = NULL, *h_exclusive = NULL;
	assert(2 == argc);
	const unsigned int N = 1<<atoi(argv[1]); // vector size
	assert(0 == N%CUDA_WARP_SIZE);

	checkCudaErrors(cudaMalloc(&d_values, N * sizeof(int)));
	checkCudaErrors(cudaMalloc(&d_inclusive, N * sizeof(int)));
	checkCudaErrors(cudaMalloc(&d_exclusive, N * sizeof(int)));
	checkCudaErrors(cudaMallocHost(&h_values, N * sizeof(int)));
	checkCudaErrors(cudaMallocHost(&h_inclusive, N * sizeof(int)));
	checkCudaErrors(cudaMallocHost(&h_exclusive, N * sizeof(int)));
	assert(d_values && h_values && d_inclusive && h_inclusive && d_exclusive && h_exclusive);

	dim3 block_size(BLOCK_SIZE);
	dim3 grid_size(DIV_CEIL(N,BLOCK_SIZE));
	set<<<grid_size, block_size>>>(N, d_values);
	getLastCudaError("set kernel failed");
	multi_scan<<<grid_size, block_size>>>((const int*)d_values, d_inclusive, d_exclusive);
	getLastCudaError("multi_scan kernel failed");
	checkCudaErrors(cudaDeviceSynchronize());

	checkCudaErrors(cudaMemcpy(h_values, d_values, N*sizeof(int), cudaMemcpyDefault));
	checkCudaErrors(cudaMemcpy(h_inclusive, d_inclusive, N*sizeof(int), cudaMemcpyDefault));
	checkCudaErrors(cudaMemcpy(h_exclusive, d_exclusive, N*sizeof(int), cudaMemcpyDefault));

	unsigned int block=0;
	unsigned int thread=0;
	for (block=0; block<N/BLOCK_SIZE; block++) {
		int inclusive = 0, exclusive = 0;
		for (thread=0; thread<BLOCK_SIZE; ++thread) {
			unsigned int i = block*BLOCK_SIZE + thread;
			inclusive += h_values[i];
			if (inclusive!=h_inclusive[i] || exclusive!=h_exclusive[i]) {
				goto EXIT;
			}
			exclusive += h_values[i];
		}
	}
EXIT:
	if (N/BLOCK_SIZE!=block) { // if error, print complete block
		printf("Error in block %d, thread %d\n", block, thread);
		for (unsigned int t=0; t<BLOCK_SIZE; ++t) {
			printf("%d ", h_values[block+t]);
		} printf("\n"); printf("--------\n");
		for (unsigned int t=0; t<BLOCK_SIZE; ++t) {
			printf("%d ", h_inclusive[block+t]);
		} printf("\n"); printf("--------\n");
		for (unsigned int t=0; t<BLOCK_SIZE; ++t) {
			printf("%d ", h_exclusive[block+t]);
		} printf("\n"); printf("--------\n");
	}

	checkCudaErrors(cudaFreeHost(h_values));
	checkCudaErrors(cudaFreeHost(h_inclusive));
	checkCudaErrors(cudaFreeHost(h_exclusive));
	checkCudaErrors(cudaFree(d_values));
	checkCudaErrors(cudaFree(d_inclusive));
	checkCudaErrors(cudaFree(d_exclusive));

	return 0;
}
